# -*- coding: utf-8 -*-

"""Basic EnergySystem class

This file is part of project oemof (github.com/oemof/oemof). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location oemof/oemof/energy_system.py

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Cord Kaldemeyer <>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>

SPDX-License-Identifier: MIT
"""

import logging
import os
import warnings
from collections import deque

import blinker
import dill as pickle

from oemof.network.groupings import DEFAULT as BY_UID
from oemof.network.groupings import Entities
from oemof.network.groupings import Grouping


class EnergySystem:
    r"""Defining an energy supply system to use oemof's solver libraries.

    Note
    ----
    The list of regions is not necessary to use the energy system with solph.

    Parameters
    ----------
    entities : list of :class:`Entity <oemof.core.network.Entity>`, optional
        A list containing the already existing :class:`Entities
        <oemof.core.network.Entity>` that should be part of the energy system.
        Stored in the :attr:`entities` attribute.
        Defaults to `[]` if not supplied.
    timeindex : pandas.datetimeindex
        Defines the time range and, if equidistant, the timeindex for the
        energy system
    timeincrement : numeric (sequence)
        Define the timeincrement for the energy system
    groupings : list
        The elements of this list are used to construct :class:`Groupings
        <oemof.core.energy_system.Grouping>` or they are used directly if they
        are instances of :class:`Grouping <oemof.core.energy_system.Grouping>`.
        These groupings are then used to aggregate the entities added to this
        energy system into :attr:`groups`.
        By default, there'll always be one group for each :attr:`uid
        <oemof.core.network.Entity.uid>` containing exactly the entity with the
        given :attr:`uid <oemof.core.network.Entity.uid>`.
        See the :ref:`examples <energy-system-examples>` for more information.

    Attributes
    ----------
    entities : list of :class:`Entity <oemof.core.network.Entity>`
        A list containing the :class:`Entities <oemof.core.network.Entity>`
        that comprise the energy system.
    groups : dict
    results : dictionary
        A dictionary holding the results produced by the energy system.
        Is `None` while no results are produced.
        Currently only set after a call to :meth:`optimize` after which it
        holds the return value of :meth:`om.results()
        <oemof.solph.optimization_model.OptimizationModel.results>`.
        See the documentation of that method for a detailed description of the
        structure of the results dictionary.
    timeindex : pandas.index, optional
        Define the time range and increment for the energy system. This is an
        optional attribute but might be import for other functions/methods that
        use the EnergySystem class as an input parameter.


    .. _energy-system-examples:
    Examples
    --------

    Regardles of additional groupings, :class:`entities
    <oemof.core.network.Entity>` will always be grouped by their :attr:`uid
    <oemof.core.network.Entity.uid>`:

    >>> from oemof.network.network import Node
    >>> es = EnergySystem()
    >>> bus = Node(label='electricity')
    >>> es.add(bus)
    >>> bus is es.groups['electricity']
    True
    >>> es.dump()  # doctest: +ELLIPSIS
    'Attributes dumped to ...
    >>> es = EnergySystem()
    >>> es.restore()  # doctest: +ELLIPSIS
    'Attributes restored from ...
    >>> bus is es.groups['electricity']
    False
    >>> es.groups['electricity']
    "<oemof.network.network.nodes.Node: 'electricity'>"

    For simple user defined groupings, you can just supply a function that
    computes a key from an :class:`entity <oemof.core.network.Entity>` and the
    resulting groups will be sets of :class:`entities
    <oemof.network.Entity>` stored under the returned keys, like in this
    example, where :class:`entities <oemof.network.Entity>` are grouped by
    their `type`:

    >>> es = EnergySystem(groupings=[type])
    >>> buses = set(Node(label="Node {}".format(i)) for i in range(9))
    >>> es.add(*buses)
    >>> class Sink(Node):
    ...     pass
    >>> components = set(Sink(label="Component {}".format(i))
    ...                   for i in range(9))
    >>> es.add(*components)
    >>> buses == es.groups[Node]
    True
    >>> components == es.groups[Sink]
    True

    """

    signals = {}
    """A dictionary of blinker_ signals emitted by energy systems.

    Currently only one signal is supported. This signal is emitted whenever a
    `node <oemof.network.Node>` is `add`ed to an energy system. The
    signal's `sender` is set to the `node <oemof.network.Node>` that got
    added to the energy system so that `node <oemof.network.Node>` have an
    easy way to only receive signals for when they themselves get added to an
    energy system.

    .. _blinker: https://blinker.readthedocs.io/en/stable/
    """

    def __init__(
        self,
        *,
        groupings=None,
        results=None,
        timeindex=None,
        timeincrement=None,
        temporal=None,
        nodes=None,
        entities=None,
    ):
        if groupings is None:
            groupings = []
        if entities is not None:
            warnings.warn(
                "Parameter 'entities' is deprecated, use 'nodes'"
                + " instead. Will overwrite nodes.",
                FutureWarning,
            )
            nodes = entities
        if nodes is None:
            nodes = []

        self._first_ungrouped_node_index_ = 0
        self._groups = {}
        self._groupings = [BY_UID] + [
            g if isinstance(g, Grouping) else Entities(g) for g in groupings
        ]
        self._nodes = {}

        self.results = results
        self.timeindex = timeindex
        self.timeincrement = timeincrement
        self.temporal = temporal
        self.add(*nodes)

    def add(self, *nodes):
        """Add :class:`nodes <oemof.network.Node>` to this energy system."""
        new_nodes = {node.label: node for node in nodes}
        if self._nodes.keys().isdisjoint(new_nodes.keys()):
            self._nodes.update(new_nodes)
        else:
            common_labels = list(self._nodes.keys() & new_nodes.keys())
            raise ValueError(
                "EnergySystem already contains Node(s) labeled: "
                + ", ".join(common_labels)
            )
        self._nodes.update(new_nodes)
        for n in nodes:
            self.signals[type(self).add].send(n, EnergySystem=self)

    signals[add] = blinker.signal(add)

    @property
    def groups(self):
        gs = self._groups
        deque(
            (
                g(n, gs)
                for g in self._groupings
                for n in list(self.nodes)[self._first_ungrouped_node_index_ :]
            ),
            maxlen=0,
        )
        self._first_ungrouped_node_index_ = len(self.nodes)
        return self._groups

    @property
    def node(self):
        return self._nodes

    @property
    def nodes(self):
        return self._nodes.values()

    def flows(self):
        return {
            (source, target): source.outputs[target]
            for source in self.nodes
            for target in source.outputs
        }

    def check(self):
        error_message = (
            "Node {n} not part of EnergySystem "
            + "but Flow ({i}, {o}) exists."
        )

        for n in self.nodes:
            for o in n.outputs.keys():
                if o not in self.nodes:
                    raise RuntimeError(error_message.format(n=n, i=n, o=o))
            for i in n.inputs.keys():
                if i not in self.nodes:
                    raise RuntimeError(error_message.format(n=n, i=i, o=n))

    # Begin: to be removed in a future version
    @staticmethod
    def _deprecated_path_handling(dpath, filename, consider_dpath):
        if consider_dpath:
            if dpath is None:
                bpath = os.path.join(os.path.expanduser("~"), ".oemof")
                if not os.path.isdir(bpath):
                    os.mkdir(bpath)
                dpath = os.path.join(bpath, "dumps")
                if not os.path.isdir(dpath):
                    os.mkdir(dpath)

                warnings.warn(
                    "Default directory for oemof dumps will change"
                    + " from ~/.oemof/dumps/ to ./ in a future version."
                    + " Set 'consider_dpath' to False to already use"
                    + " the new default.",
                    FutureWarning,
                )
            else:
                warnings.warn(
                    "Parameter 'dpath' will be removed in a future"
                    + " version. You can give the directory as part"
                    + " of the filename and set 'consider_dpath' to"
                    + " False to suppress this waring.",
                    FutureWarning,
                )
            if filename is None:
                filename = "es_dump.oemof"

            filename = os.path.join(dpath, filename)
        else:
            if dpath is not None:
                if filename is None:
                    # Interpret dpath as intended to be filename,
                    # as it might be given as positional argument.
                    filename = dpath
                else:
                    raise ValueError(
                        "You set filename and dpath but told that"
                        + " dpath should be ignored."
                    )

        return filename
        # End: to be removed in a future version

    def dump(
        self,
        dpath=None,  # to be removed in a future version
        filename=None,
        consider_dpath=True,  # to be removed in a future version
    ):
        """Dump an EnergySystem instance.

        Parameters
        ----------
        dpath : str
            Path to write your dump in.
        filename : str
            Filename to write your dump to.
        consider_dpath : bool
            Use separate parameters for path (default: ~/.oemof/) and filename.
        """
        # Start: to be removed in a future version
        filename = self._deprecated_path_handling(
            dpath, filename, consider_dpath
        )
        # End: to be removed in a future version

        pickle.dump(self.__dict__, open(filename, "wb"))

        msg = f"Attributes dumped to {filename}."
        logging.debug(msg)
        return msg

    def restore(
        self,
        dpath=None,  # to be removed in a future version
        filename=None,
        consider_dpath=True,  # to be removed in a future version
    ):
        """Restore an EnergySystem instance.

        Parameters
        ----------
        dpath : str
            Path to write your dump in.
        filename : str
            Filename to write your dump to.
        consider_dpath : bool
            Use separate parameters for path (defualt: ~/.oemof/) and filename.
        """
        logging.info(
            "Restoring attributes will overwrite existing attributes."
        )
        # Start: to be removed in a future version
        filename = self._deprecated_path_handling(
            dpath, filename, consider_dpath
        )
        # End: to be removed in a future version

        self.__dict__ = pickle.load(open(filename, "rb"))

        msg = f"Attributes restored from {filename}."
        logging.debug(msg)
        return msg



# -*- coding: utf-8 -*-

"""This package contains the abstract entity classes used to model
energy systems.

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Cord Kaldemeyer <>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>

SPDX-License-Identifier: MIT
"""

from functools import total_ordering


@total_ordering
class Entity:
    """Represents an Entity in an energy system graph.

    Abstract superclass of the general types of entities of an energy system
    graph, collecting attributes and operations common to all types of nodes.
    Users should neither instantiate nor subclass this, but use
    :class:`Component`, :class:`Bus`, :class:`Edge` or one of their subclasses
    instead.

    .. role:: python(code)
      :language: python

    Parameters
    ----------
    label: `hashable`, optional
        Used as the string representation of this node. If this parameter is
        not an instance of :class:`str` it will be converted to a string and
        the result will be used as this node's :attr:`label`, which should be
        unique with respect to the other nodes in the energy system graph this
        node belongs to. If this parameter is not supplied, the string
        representation of this node will instead be generated based on this
        nodes `class` and `id`.

    custom_properties: `dict`
        This dictionary that can be used to store information that can be used
        to easily attach custom information to any Entity.
    """

    def __init__(self, label, *, custom_properties=None):
        self._label = label
        if custom_properties is None:
            custom_properties = {}
        self.custom_properties = custom_properties

    def __eq__(self, other):
        return id(self) == id(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __hash__(self):
        return hash(self.label)

    def __str__(self):
        return str(self.label)

    def __repr__(self):
        return repr(
            "<{0.__module__}.{0.__name__}: {1!r}>".format(
                type(self), self.label
            )
        )

    @property
    def label(self):
        """
        If this node was given a `label` on construction, this
        attribute holds the actual object passed as a parameter. Otherwise
        `node.label` is a synonym for `str(node)`.
        """
        try:
            return self._label if self._label is not None else self._id_label
        except AttributeError:  # Workaround for problems with pickle/dill
            return hash(self._id_label)

    @property
    def _id_label(self):
        return "<{} #0x{:x}>".format(type(self).__name__, id(self))



# -*- coding: utf-8 -*-

"""All you need to create groups of stuff in your energy system.

This file is part of project oemof (github.com/oemof/oemof). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location oemof/oemof/groupings.py

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>

SPDX-License-Identifier: MIT
"""

from collections.abc import Hashable
from collections.abc import Iterable
from collections.abc import Mapping
from collections.abc import MutableMapping
from itertools import chain
from itertools import filterfalse

from oemof.network.network.edge import Edge

# TODO: Update docstrings.
#
#   * Make them easier to understand.
#   * Update them to use nodes instead of entities.
#


class Grouping:
    """
    Used to aggregate :class:`entities <oemof.network.Entity>` in an
    :class:`energy system <oemof.energy_system.EnergySystem>` into
    :attr:`groups <oemof.energy_system.EnergySystem.groups>`.

    The way :class:`Groupings <Grouping>` work is that each :class:`Grouping`
    :obj:`g` of an energy system is called whenever an :class:`entity
    <oemof.network.Entity>` is added to the energy system (and for each
    :class:`entity <oemof.network.Entity>` already present, if the energy
    system is created with existing enties).
    The call :obj:`g(e, groups)`, where :obj:`e` is an :class:`entity
    <oemof.network.Entity>` and :attr:`groups
    <oemof.energy_system.EnergySystem.groups>` is a dictionary mapping
    group keys to groups, then uses the three functions :meth:`key
    <Grouping.key>`, :meth:`value <Grouping.value>` and :meth:`merge
    <Grouping.merge>` in the following way:

        - :meth:`key(e) <Grouping.key>` is called to obtain a key :obj:`k`
          under which the group should be stored,
        - :meth:`value(e) <Grouping.value>` is called to obtain a value
          :obj:`v` (the actual group) to store under :obj:`k`,
        - if you supplied a :func:`filter` argument, :obj:`v` is
          :func:`filtered <builtins.filter>` using that function,
        - otherwise, if there is not yet anything stored under
          :obj:`groups[k]`, :obj:`groups[k]` is set to :obj:`v`. Otherwise
          :meth:`merge <Grouping.merge>` is used to figure out how to merge
          :obj:`v` into the old value of :obj:`groups[k]`, i.e.
          :obj:`groups[k]` is set to :meth:`merge(v, groups[k])
          <Grouping.merge>`.

    Instead of trying to use this class directly, have a look at its
    subclasses, like :class:`Entities`, which should cater for most use cases.

    Notes
    -----

    When overriding methods using any of the constructor parameters, you don't
    have access to :obj:`self` in the corresponding function. If you need
    access to :obj:`self`, subclass :class:`Grouping` and override the methods
    in the subclass.

    A :class:`Grouping` may be called more than once on the same object
    :obj:`e`, so one should make sure that user defined :class:`Grouping`
    :obj:`g` is idempotent, i.e. :obj:`g(e, g(e, d)) == g(e, d)`.

    Parameters
    ----------

    key: callable or hashable

        Specifies (if not callable) or extracts (if callable) a :meth:`key
        <Grouping.key>` for each :class:`entity <oemof.network.Entity>` of
        the :class:`energy system <oemof.energy_system.EnergySystem>`.

    constant_key: hashable (optional)

        Specifies a constant :meth:`key <Grouping.key>`. Keys specified using
        this parameter are not called but taken as is.

    value: callable, optional

        Overrides the default behaviour of :meth:`value <Grouping.value>`.

    filter: callable, optional

        If supplied, whatever is returned by :meth:`value` is :func:`filtered
        <builtins.filter>` through this. Mostly useful in conjunction with
        static (i.e. non-callable) :meth:`keys <key>`.
        See :meth:`filter` for more details.

    merge: callable, optional

        Overrides the default behaviour of :meth:`merge <Grouping.merge>`.

    """

    def __init__(self, key=None, constant_key=None, filter=None, **kwargs):
        if key and constant_key:
            raise TypeError(
                "Grouping arguments `key` and `constant_key` are "
                + " mutually exclusive."
            )
        if constant_key:
            self.key = lambda _: constant_key
        elif key:
            self.key = key
        else:
            raise TypeError(
                "Grouping constructor missing required argument: "
                + "one of `key` or `constant_key`."
            )
        self.filter = filter
        for kw in ["value", "merge"]:
            if kw in kwargs:
                setattr(self, kw, kwargs[kw])

    def key(self, node):
        """Obtain a key under which to store the group.

        You have to supply this method yourself using the :obj:`key` parameter
        when creating :class:`Grouping` instances.

        Called for every :class:`node <oemof.network.Entity>` of the
        energy system. Expected to return the key (i.e. a valid
        :class:`hashable`) under which the group
        :meth:`value(node) <Grouping.value>` will be
        stored. If it should be added to more than one group, return a
        :class:`list` (or any other non-:class:`hashable <Hashable>`,
        :class:`iterable`) containing the group keys.

        Return :obj:`None` if you don't want to store :obj:`e` in a group.
        """
        raise NotImplementedError(
            "\n\n"
            "There is no default implementation for `Groupings.key`.\n"
            "Congratulations, you managed to execute supposedly "
            "unreachable code.\n"
            "Please let us know by filing a bug at:\n\n    "
            "https://github.com/oemof/oemof/issues\n"
        )

    def value(self, e):
        """Generate the group obtained from :obj:`e`.

        This methd returns the actual group obtained from :obj:`e`. Like
        :meth:`key <Grouping.key>`, it is called for every :obj:`e` in the
        energy system. If there is no group stored under :meth:`key(e)
        <Grouping.key>`, :obj:`groups[key(e)]` is set to :meth:`value(e)
        <Grouping.value>`. Otherwise :meth:`merge(value(e), groups[key(e)])
        <Grouping.merge>` is called.

        The default returns the :class:`entity <oemof.network.Entity>`
        itself.
        """
        return e

    def merge(self, new, old):
        """Merge a known :obj:`old` group with a :obj:`new` one.

        This method is called if there is already a value stored under
        :obj:`group[key(e)]`. In that case, :meth:`merge(value(e),
        group[key(e)]) <Grouping.merge>` is called and should return the new
        group to store under :meth:`key(e) <Grouping.key>`.

        The default behaviour is to raise an error if :obj:`new` and :obj:`old`
        are not identical.
        """
        if old is new:
            return old
        raise ValueError(
            "\nGrouping \n  "
            "{}:{}\nand\n  {}:{}\ncollides.\n".format(
                id(old), old, id(new), new
            )
            + "Possibly duplicate uids/labels?"
        )

    def filter(self, group):
        """
        :func:`Filter <builtins.filter>` the group returned by :meth:`value`
        before storing it.

        Should return a boolean value. If the :obj:`group` returned by
        :meth:`value` is :class:`iterable <collections.abc.Iterable>`, this
        function is used (via Python's :func:`builtin filter
        <builtins.filter>`) to select the values which should be retained in
        :obj:`group`. If :obj:`group` is not :class:`iterable
        <collections.abc.Iterable>`, it is simply called on :obj:`group` itself
        and the return value decides whether :obj:`group` is stored
        (:obj:`True`) or not (:obj:`False`).

        """
        raise NotImplementedError(
            "\n\n"
            "`Groupings.filter` called without being overridden.\n"
            "Congratulations, you managed to execute supposedly "
            "unreachable code.\n"
            "Please let us know by filing a bug at:\n\n    "
            "https://github.com/oemof/oemof/issues\n"
        )

    def __call__(self, e, d):
        k = self.key(e) if callable(self.key) else self.key
        if k is None:
            return
        v = self.value(e)
        if isinstance(v, MutableMapping):
            for k in list(filterfalse(self.filter, v)):
                v.pop(k)
        elif isinstance(v, Mapping):
            v = type(v)(dict((k, v[k]) for k in filter(self.filter, v)))
        elif isinstance(v, Iterable):
            v = type(v)(filter(self.filter, v))
        elif self.filter and not self.filter(v):
            return
        if not v:
            return
        for group in (
            k
            if (isinstance(k, Iterable) and not isinstance(k, Hashable))
            else [k]
        ):
            d[group] = self.merge(v, d[group]) if group in d else v


class Entities(Grouping):
    """
    Specialises :class:`Grouping` to group
    :class:`entities <oemof.network.Entity>`
    into :class:`sets <set>`.
    """

    def value(self, e):
        """
        Returns a :class:`set` containing only :obj:`e`, so groups are
        :class:`sets <set>` of :class:`entity <oemof.network.Entity>`.
        """
        return {e}

    def merge(self, new, old):
        """
        :meth:`Updates <set.update>` :obj:`old` to be the union of :obj:`old`
        and :obj:`new`.
        """
        return old.union(new)


class Flows(Entities):
    """
    Specialises :class:`Grouping` to group the flows connected to :class:`nodes
    <oemof.network.Entity>` into :class:`sets <set>`.
    Note that this specifically means that the :meth:`key <Flows.key>`, and
    :meth:`value <Flows.value>` functions act on a set of flows.
    """

    def value(self, flows):
        """
        Returns a :class:`set` containing only :obj:`flows`, so groups are
        :class:`sets <set>` of flows.
        """
        return set(flows)

    def __call__(self, n, d):
        flows = (
            {n}
            if isinstance(n, Edge)
            else set(chain(n.outputs.values(), n.inputs.values()))
        )
        super().__call__(flows, d)


class FlowsWithNodes(Entities):
    """
    Specialises :class:`Grouping` to act on the flows connected to
    :class:`nodes <oemof.network.Node>` and create :class:`sets <set>` of
    :obj:`(source, target, flow)` tuples.
    Note that this specifically means that the :meth:`key <Flows.key>`, and
    :meth:`value <Flows.value>` functions act on sets like these.
    """

    def value(self, tuples):
        """
        Returns a :class:`set` containing only :obj:`tuples`, so groups are
        :class:`sets <set>` of :obj:`tuples`.
        """
        return set(tuples)

    def __call__(self, n, d):
        tuples = (
            {(n.input, n.output, n)}
            if isinstance(n, Edge)
            else set(
                chain(
                    ((n, t, f) for (t, f) in n.outputs.items()),
                    ((s, n, f) for (s, f) in n.inputs.items()),
                )
            )
        )
        super().__call__(tuples, d)


def _uid_or_str(node_or_entity):
    """
    Helper function to support the transition from `Entitie`s to `Entity`s.
    """
    return (
        node_or_entity.uid
        if hasattr(node_or_entity, "uid")
        else str(node_or_entity)
    )


DEFAULT = Grouping(_uid_or_str)
"""The default :class:`Grouping`.

This one is always present in an :class:`energy system
<oemof.energy_system.EnergySystem>`. It stores every :class:`entity
<oemof.network.Entity>` under its :attr:`uid
<oemof.network.Entity.uid>` and raises an error if another :class:`entity
<oemof.network.Entity>` with the same :attr:`uid
<oemof.network.Entity.uid>` get's added to the :class:`energy system
<oemof.energy_system.EnergySystem>`.
"""



from oemof.network import Node


class TestNode:
    """Tests for Node class"""

    def test_init_basic(self):
        """Basic initialisation"""
        subnet = Node("test_subnet")

        assert subnet.label == "test_subnet"
        assert subnet.parent is None
        assert subnet.depth == 0
        assert len(subnet.subnodes) == 0

    def test_init_with_parent_node(self):
        """Initialisation with parent_node"""
        parent = Node("parent")
        child = Node("child", parent_node=parent)

        assert child.parent == parent
        assert child.depth == 1

    def test_init_with_custom_properties(self):
        """Initialisation with custom_properties"""
        props = {"custom_key": "custom_value"}
        node = Node("test", custom_properties=props)

        assert node.custom_properties == props

    def test_subnode_addition(self):
        """Add existing Node"""
        subnet = Node("parent")

        # add single Node
        subnode = Node("child")
        subnet.add(subnode)

        assert isinstance(subnet.subnodes, tuple)
        assert len(subnet.subnodes) == 1
        assert subnet.subnodes[0] == subnode
        assert subnode.parent == subnet
        assert subnode.depth == 1
        assert subnode.label == "child"

        # add multiple Nodes
        subnode2 = Node("child2")
        subnode3 = Node("child3")
        subnet.add(subnode2, subnode3)

        assert isinstance(subnet.subnodes, tuple)
        assert len(subnet.subnodes) == 3

        # labels are unchanged
        assert subnode2.label == "child2"
        assert subnode3.label == "child3"

        # hierachy properties are adjusted
        assert subnode2.parent == subnet
        assert subnode3.parent == subnet
        assert subnode2.depth == 1
        assert subnode3.depth == 1

    def test_subnode_creation(self):
        """Create Subnode with subnode() method"""
        subnet = Node("parent")

        subnode = subnet.subnode(Node, "child")

        assert isinstance(subnet.subnodes, tuple)
        assert len(subnet.subnodes) == 1
        assert subnet.subnodes[0] == subnode
        assert isinstance(subnode, Node)
        assert subnode.parent == subnet
        assert subnode.depth == 1
        assert subnode.label == ("child", "parent")

    def test_subnode_nested_tuples(self):
        """Add Subnode with subnode() method and tuples as labels"""
        subnet = Node(("parent", "electricity"))

        subnode = subnet.subnode(Node, ("child", "electricity"))

        assert isinstance(subnet.subnodes, tuple)
        assert len(subnet.subnodes) == 1
        assert subnet.subnodes[0] == subnode
        assert isinstance(subnode, Node)
        assert subnode.parent == subnet
        assert subnode.depth == 1
        assert subnode.label == (
            ("child", "electricity"),
            ("parent", "electricity"),
        )

    def test_subnode_with_args_kwargs(self):
        """Subnode creation with extra arguments"""
        subnet = Node("parent")
        custom_props = {"test": "value"}

        subnode = subnet.subnode(Node, "child", custom_properties=custom_props)

        assert subnode.custom_properties == custom_props
        assert subnode.parent == subnet

    def test_multiple_subnodes(self):
        """Create many Subnodes"""
        subnet = Node("parent")

        child1 = subnet.subnode(Node, "child1")
        child2 = subnet.subnode(Node, "child2")
        child3 = subnet.subnode(Node, "child_subnet")

        assert len(subnet.subnodes) == 3
        assert child1 in subnet.subnodes
        assert child2 in subnet.subnodes
        assert child3 in subnet.subnodes

        # Alle sollten subnet als parent haben
        for child in subnet.subnodes:
            assert child.parent == subnet

    def test_nested_subnets(self):
        """Nested Nodes"""
        root = Node("root")
        level1 = root.subnode(Node, "level1")
        level2 = level1.subnode(Node, "level2")
        leaf = level2.subnode(Node, "leaf")

        assert root.depth == 0
        assert level1.depth == 1
        assert level2.depth == 2
        assert leaf.depth == 3

        assert leaf.label == ("leaf", "level2", "level1", "root")

    def test_complex_hierarchy(self):
        """Complex hierarchical structure"""
        root = Node("sub_energy_system")

        power_sector = root.subnode(Node, "power")
        heat_sector = root.subnode(Node, "heat")

        coal_plant = power_sector.subnode(Node, "coal_plant")

        heat_pump = heat_sector.subnode(Node, "heat_pump")

        # Validiere Struktur
        assert len(root.subnodes) == 2
        assert len(power_sector.subnodes) == 1
        assert len(heat_sector.subnodes) == 1

        # Validiere Tiefen
        assert root.depth == 0
        assert power_sector.depth == 1
        assert coal_plant.depth == 2

        # Validiere flat_labels
        assert coal_plant.label == (
            "coal_plant",
            "power",
            "sub_energy_system",
        )
        assert heat_pump.label == (
            "heat_pump",
            "heat",
            "sub_energy_system",
        )



# -*- coding: utf-8 -

"""Test the created constraints against approved constraints.

This file is part of project oemof (github.com/oemof/oemof). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location oemof/tests/test_network_classes.py

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Cord Kaldemeyer <>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>

SPDX-License-Identifier: MIT
"""

from traceback import format_exception_only as feo

import pytest

from oemof.network.energy_system import EnergySystem
from oemof.network.network import Bus
from oemof.network.network import Sink
from oemof.network.network import Source
from oemof.network.network import Transformer
from oemof.network.network.edge import Edge
from oemof.network.network.entity import Entity
from oemof.network.network.nodes import Node


class TestsNode:
    def setup_method(self):
        self.energysystem = EnergySystem()

    def test_entity_initialisation(self):
        entity = Entity(label="foo")
        assert entity.label == "foo"

    def test_symmetric_input_output_assignment(self):
        n1 = Node(label="<N1>")

        n2 = Node(label="<From N1>", inputs=[n1])
        assert n1 in n2.inputs, (
            "{0} not in {1}.inputs, ".format(n1, n2)
            + "although it should be by construction."
        )

        assert (
            n2 in n1.outputs
        ), "{0} in {1}.inputs but {1} not in {0}.outputs.".format(n1, n2)

        n3 = Node(label="<To N1>", outputs=[n1])
        assert n1 in n3.outputs, (
            "{0} not in {1}.outputs, ".format(n1, n3)
            + "although it should be by construction.",
        )
        assert n3 in n1.inputs, (
            "{0} in {1}.outputs but {1} not in {0}.inputs.".format(n1, n3),
        )

    def test_accessing_outputs_of_a_node_without_output_flows(self):
        n = Node(label="node")
        exception = None
        outputs = None
        try:
            outputs = n.outputs
        except Exception as e:
            exception = e
        assert exception is None, (
            "\n  Test accessing `outputs` on {} having no outputs.".format(n)
            + "\n  Got unexpected exception:\n"
            + "\n      {}".format(feo(type(exception), exception)[0]),
        )
        assert len(outputs) == 0, (
            "\n  Failure when testing `len(outputs)`."
            + "\n  Expected: 0."
            + "\n  Got     : {}".format(len(outputs)),
        )

    def test_accessing_inputs_of_a_node_without_input_flows(self):
        n = Node(label="node")
        exception = None
        inputs = None
        try:
            inputs = n.inputs
        except Exception as e:
            exception = e
        assert exception is None, (
            "\n  Test accessing `inputs` on {} having no inputs.".format(n)
            + "\n  Got unexpected exception:\n"
            + "\n      {}".format(feo(type(exception), exception)[0]),
        )
        assert len(inputs) == 0, (
            "\n  Failure when testing `len(inputs)`."
            + "\n  Expected: 0."
            + "\n  Got     : {}".format(len(inputs)),
        )

    def test_that_the_outputs_attribute_of_a_node_is_a_mapping(self):
        n = Node(label="node")
        exception = None
        try:
            n.outputs.values()
        except AttributeError as e:
            exception = e
        assert exception is None, (
            "\n  Test accessing `outputs.values()`"
            + " on {} having no inputs.".format(n)
            + "\n  Got unexpected exception:\n"
            + "\n      {}".format(feo(type(exception), exception)[0]),
        )

    def test_that_nodes_do_not_get_undead_flows(self):
        """Newly created nodes should only have flows assigned to them.

        A new node `n`, which re-used a previously used label `l`, retained the
        flows of those nodes which where labeled `l` before `n`. This incorrect
        behaviour is a problem if somebody wants to use different nodes with
        the same label in multiple energy systems. While this feature currently
        isn't used, it also lead to weird behaviour when running tests.

        This test ensures that new nodes only have those flows which are
        assigned to them on construction.
        """
        flow = object()
        old = Node(label="A reused label")
        bus = Node(label="bus", inputs={old: flow})
        assert bus.inputs[old].flow == flow, (
            ("\n  Expected: {0}" + "\n  Got     : {1} instead").format(
                flow, bus.inputs[old].flow
            ),
        )
        assert old.outputs[bus].flow == flow, (
            ("\n  Expected: {}" + "\n  Got     : {} instead").format(
                flow, old.outputs[bus].flow
            ),
        )
        new = Node(label="A reused label")
        assert new.outputs == {}, (
            "\n  Expected an empty dictionary of outputs."
            + "\n  Got: {} instead".format(new.outputs),
        )

    def test_modifying_outputs_after_construction(self):
        """One should be able to add and delete outputs of a node."""
        node = Node("N1")
        bus = Node("N2")
        flow = "flow"
        assert node.outputs == {}, (
            "\n  Expected an empty dictionary of outputs."
            + "\n  Got: {} (== {}) instead".format(
                node.outputs, dict(node.outputs)
            ),
        )
        node.outputs[bus] = flow
        assert node.outputs == {bus: flow}, (
            "\n  Expected {} as `node.outputs`."
            + "\n  Got    : {} (== {}) instead"
        ).format({bus: flow}, node.outputs, dict(node.outputs))

        assert node.outputs[bus] == flow, (
            "\n  Expected {} as `node.outputs[bus]`."
            + "\n  Got    : {} instead"
        ).format(flow, node.outputs[bus])

        del node.outputs[bus]
        assert node.outputs == {}, (
            "\n  Expected an empty dictionary of outputs."
            + "\n  Got: {} (== {}) instead"
        ).format(node.outputs, dict(node.outputs))

    def test_modifying_inputs_after_construction(self):
        """One should be able to add and delete inputs of a node."""
        node = Node("N1")
        bus = Node("N2")
        flow = "flow"

        assert node.inputs == {}, (
            "\n  Expected an empty dictionary of inputs."
            + "\n  Got: {} (== {}) instead"
        ).format(node.inputs, dict(node.inputs))

        node.inputs[bus] = flow
        assert node.inputs == {bus: flow}, (
            "\n  Expected {} as `node.inputs`."
            + "\n  Got    : {} (== {}) instead"
        ).format({bus: flow}, node.inputs, dict(node.inputs))
        assert node.inputs[bus] == flow, (
            "\n  Expected {} as `node.inputs[bus]`."
            + "\n  Got    : {} instead"
        ).format(flow, node.inputs[bus])
        del node.inputs[bus]
        assert node.inputs == {}, (
            "\n  Expected an empty dictionary of inputs."
            + "\n  Got: {} (== {}) instead"
        ).format(node.inputs, dict(node.inputs))

    def test_output_input_symmetry_after_modification(self):
        n1 = Node("N1")
        n2 = Node("N2")
        flow = "flow"

        n1.outputs[n2] = flow
        assert n2.inputs == {n1: flow}

    def test_input_output_symmetry_after_modification(self):
        n1 = Node("N1")
        n2 = Node("N2")
        flow = "flow"

        n1.inputs[n2] = flow
        assert n2.outputs == {n1: flow}

    def test_updating_inputs(self):
        n1 = Node("N1")
        n2 = Node("N2")
        n1n2 = "n1n2"

        n2.inputs.update({n1: n1n2})
        assert n2.inputs == {n1: n1n2}
        assert n2.inputs[n1] == n1n2
        assert n1.outputs == {n2: n1n2}
        assert n1.outputs[n2] == n1n2

    def test_updating_outputs(self):
        n1 = Node("N1")
        n2 = Node("N2")
        n1n2 = "n1n2"

        n1.outputs.update({n2: n1n2})
        assert n2.inputs == {n1: n1n2}
        assert n2.inputs[n1] == n1n2
        assert n1.outputs == {n2: n1n2}
        assert n1.outputs[n2] == n1n2

    def test_error_for_duplicate_label_argument(self):
        """
        `Node.__init__` should fail if positional and keyword args collide.
        """
        with pytest.raises(TypeError):
            Node("Positional Label", label="Keyword Label")

    def test_entity_input_output_type_assertions(self):
        """
        `'Node'` should only accept `Node` instances
        as input/output targets.
        """
        with pytest.raises(ValueError):
            Node("An entity with an output", outputs={"Not a Node": "A Flow"})

        with pytest.raises(ValueError):
            Node("An entity with an input", inputs={"Not a Node": "A Flow"})

    def test_node_requires_label(self):
        """
        A `Node` without `label` cannot be constructed.
        """
        with pytest.raises(TypeError):
            _ = Node()

    def test_node_label_if_its_not_explicitly_specified(self):
        """If not explicitly given, a `Node`'s label is based on its `id`."""
        n = Node(label=None)
        assert "0x{:x}>".format(id(n)) in n.label

    def test_depth_default(self):
        n = Node(label="node")
        assert n.depth == 0


class TestsEdge:
    def setup_method(self):
        pass

    def test_edge_construction_side_effects(self):
        """Constructing an `Edge` should affect it's input/output `Node`s.

        When constructing an `Edge`, the `inputs` and `outputs` of its output
        and input `Node`s should be set appropriately.
        """
        source = Node(label="source")
        target = Node(label="target")
        edge = Edge(input_node=source, output_node=target)
        assert target in source.outputs, (
            "{} not in {} after constructing {}.".format(
                target, source.outputs, edge
            ),
        )
        assert source in target.inputs, (
            "{} not in {} after constructing {}.".format(
                source, target.outputs, edge
            ),
        )

    def test_label_as_positional_argument(self):
        o = object()
        n = Node(o)
        assert n.label is o, (
            "Setting `label` as positional parameter argument failed."
            "\n  Expected: {!r}"
            "\n  Got     : {!r}"
        ).format(o, n.label)

    def test_edge_failure_for_colliding_arguments(self):
        """
        `Edge` initialisation fails when colliding arguments are supplied.
        """
        with pytest.raises(ValueError):
            Edge(flow=object(), values=object())

    def test_alternative_edge_construction_from_mapping(self):
        """`Edge.from_object` treats mappings as keyword arguments."""
        i, o, f = (Node("input"), Node("output"), "flow")
        with pytest.raises(ValueError):
            Edge.from_object({"flow": i, "values": o})
        edge = Edge.from_object({"input_node": i, "output_node": o, "flow": f})
        assert edge.input == i
        assert edge.output == o
        assert edge.values == f
        assert edge.flow == f

    def test_flow_setter(self):
        """`Edge.flow`'s setter relays to `values`."""
        e = Edge(values="initial values")
        assert e.flow == "initial values"
        assert e.values == "initial values"
        e.flow = "new values set via `e.flow`"
        assert e.flow == "new values set via `e.flow`"
        assert e.values == "new values set via `e.flow`"


class TestsEnergySystemNodesIntegration:
    def setup_method(self):
        self.es = EnergySystem()

    def test_entity_registration(self):
        n1 = Node(label="<B1>")
        self.es.add(n1)
        assert self.es.node["<B1>"] == n1

        n2 = Node(label="<B2>")

        n3 = Node(label="<TF1>", inputs=[n1], outputs=[n2])
        self.es.add(n3, n2)
        assert self.es.node["<B2>"] == n2
        assert self.es.node["<TF1>"] == n3


def test_deprecated_classes():
    with pytest.warns(FutureWarning):
        Bus("bus")
    with pytest.warns(FutureWarning):
        Sink("sink")
    with pytest.warns(FutureWarning):
        Source("source")
    with pytest.warns(FutureWarning):
        Transformer("transformer")


def test_custom_properties():
    node0 = Node("n0")

    assert not node0.custom_properties

    node1 = Node(
        "n1",
        custom_properties={
            "foo": "bar",
            1: 2,
        },
    )
    assert node1.custom_properties["foo"] == "bar"
    assert node1.custom_properties[1] == 2


def test_comparision():
    node0 = Node(label=0)
    node1 = Node(label=2)
    node2 = Node(label=-5)

    assert node0 < node1
    assert node0 > node2



# -*- coding: utf-8 -

"""Basic tests.

This file is part of project oemof.network (github.com/oemof/oemof-network).

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Cord Kaldemeyer <>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>

SPDX-License-Identifier: MIT
"""
from collections.abc import Iterable
from pprint import pformat

from oemof.network import energy_system as es
from oemof.network.groupings import Entities
from oemof.network.groupings import Grouping
from oemof.network.network.nodes import Node


def test_entity_grouping_on_construction():
    bus = Node(label="test bus")
    ensys = es.EnergySystem(nodes=[bus])
    assert ensys.groups[bus.label] is bus


def test_that_none_is_not_a_valid_group():
    def by_uid(n):
        if "Not in 'Group'" in n.uid:
            return None
        else:
            return "Group"

    ensys = es.EnergySystem(groupings=[by_uid])

    ungrouped = [Node(label="Not in 'Group': {}".format(i)) for i in range(10)]
    grouped = [Node(label="In 'Group': {}".format(i)) for i in range(10)]
    assert None not in ensys.groups
    for g in ensys.groups.values():
        for e in ungrouped:
            if isinstance(g, Iterable) and not isinstance(g, str):
                assert e not in g
        for e in grouped:
            if isinstance(g, Iterable) and not isinstance(g, str):
                assert e in g


def test_defining_multiple_groupings_with_one_function():
    def assign_to_multiple_groups_in_one_go(n):
        g1 = n.label[-1]
        g2 = n.label[0:3]
        return [g1, g2]

    ensy = es.EnergySystem(groupings=[assign_to_multiple_groups_in_one_go])
    nodes = [
        Node(
            label=("Foo: " if i % 2 == 0 else "Bar: ")
            + "{}".format(i)
            + ("A" if i < 5 else "B")
        )
        for i in range(10)
    ]
    ensy.add(*nodes)
    for group in ["Foo", "Bar", "A", "B"]:
        assert len(ensy.groups[group]) == 5, (
            "\n  Failed testing length of group '{}'."
            + "\n  Expected: 5"
            + "\n  Got     : {}"
            + "\n  Group   : {}"
        ).format(
            group,
            len(ensy.groups[group]),
            sorted([e.label for e in ensy.groups[group]]),
        )


def test_grouping_filter_parameter():
    g1 = Grouping(
        key=lambda e: "The Special One",
        filter=lambda e: "special" in str(e),
    )
    g2 = Entities(
        key=lambda e: "A Subset", filter=lambda e: "subset" in str(e)
    )
    ensys = es.EnergySystem(groupings=[g1, g2])
    special = Node(label="special")
    subset = set(Node(label="subset: {}".format(i)) for i in range(10))
    others = set(Node(label="other: {}".format(i)) for i in range(10))
    ensys.add(special, *subset)
    ensys.add(*others)
    assert ensys.groups["The Special One"] == special
    assert ensys.groups["A Subset"] == subset


def test_proper_filtering():
    """`Grouping.filter` should not be "all or nothing".

    There was a bug where, if `Grouping.filter` returned `False` only for
    some elements of `Grouping.value(e)`, those elements where actually
    retained.
    This test makes sure that the bug doesn't resurface again.
    """
    g = Entities(
        key="group",
        value=lambda _: {1, 2, 3, 4},
        filter=lambda x: x % 2 == 0,
    )
    ensys = es.EnergySystem(groupings=[g])
    special = Node(label="object")
    ensys.add(special)
    assert ensys.groups["group"] == {2, 4}


def test_non_callable_group_keys():
    collect_everything = Entities(key="everything")
    g1 = Grouping(key="The Special One", filter=lambda e: "special" in e.label)
    g2 = Entities(key="A Subset", filter=lambda e: "subset" in e.label)
    ensys = es.EnergySystem(groupings=[g1, g2, collect_everything])
    special = Node(label="special")
    subset = set(Node(label="subset: {}".format(i)) for i in range(2))
    others = set(Node(label="other: {}".format(i)) for i in range(2))
    everything = subset.union(others)
    everything.add(special)
    ensys.add(*everything)
    assert ensys.groups["The Special One"] == special
    assert ensys.groups["A Subset"] == subset
    assert ensys.groups["everything"] == everything


def test_grouping_laziness():
    """Energy system `groups` should be fully lazy.

    `Node`s added to an energy system should only be tested for and put
    into their respective groups right before the `groups` property of an
    energy system is accessed.
    """
    group = "Group"
    g = Entities(key=group, filter=lambda n: getattr(n, "group", False))
    ensys = es.EnergySystem(groupings=[g])
    buses = [Node("Grouped"), Node("Ungrouped one"), Node("Ungrouped two")]
    ensys.add(buses[0])
    buses[0].group = True
    ensys.add(*buses[1:])
    assert group in ensys.groups, (
        (
            "\nExpected to find\n\n  `{!r}`\n\n"
            "in `es.groups`.\nGot:\n\n  `{}`"
        ).format(
            group,
            "\n   ".join(pformat(set(ensys.groups.keys())).split("\n")),
        ),
    )
    assert buses[0] in ensys.groups[group], (
        "\nExpected\n\n  `{}`\n\nin `es.groups['{}']`:\n\n  `{}`".format(
            "\n   ".join(pformat(buses[0]).split("\n")),
            group,
            "\n   ".join(pformat(ensys.groups[group]).split("\n")),
        ),
    )


def test_constant_group_keys():
    """Callable keys passed in as `constant_key` should not be called.

    The `constant_key` parameter can be used to specify callable group keys
    without having to worry about `Grouping`s trying to call them. This
    test makes sure that the parameter is handled correctly.
    """

    def everything():
        return "everything"

    collect_everything = Entities(constant_key=everything)
    ensys = es.EnergySystem(groupings=[collect_everything])
    node = Node(label="A Node")
    ensys.add(node)
    assert "everything" not in ensys.groups
    assert everything in ensys.groups
    assert ensys.groups[everything] == {node}
    assert everything() == "everything"



# -*- coding: utf-8 -

"""Tests for the EnergySystem class.

This file is part of project oemof.network (github.com/oemof/oemof-network).

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Cord Kaldemeyer <>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>
SPDX-FileCopyrightText: Pierre-Francois Duc <pierre-francois@rl-institut.de>

SPDX-License-Identifier: MIT
"""

from pathlib import Path

import pytest
from networkx import DiGraph

from oemof.network import Edge
from oemof.network import Node
from oemof.network import graph
from oemof.network.energy_system import EnergySystem


def test_ensys_init():
    node = Node("label")
    ensys = EnergySystem(nodes=[node])
    assert node in ensys.nodes

    with pytest.warns(FutureWarning):
        ensys = EnergySystem(entities=[node])
        assert node in ensys.nodes


class TestDumpRestore:
    def setup_method(self):
        self.es = EnergySystem()

        node0 = Node(label="node0")
        node1 = Node(label="node1", inputs={node0: Edge()})
        self.es.add(node0, node1)

    def test_dump_restore_cwd(self):
        """Test dumping and restoring to and from current working directory."""

        filename = "./es_test_dump.oemof"

        msg = self.es.dump(
            filename=filename,
            consider_dpath=False,
        )

        assert filename in msg

        es = EnergySystem()
        msg = es.restore(
            filename=filename,
            consider_dpath=False,
        )

        assert filename in msg
        assert len(es.nodes) == 2
        assert isinstance(es.node["node0"], Node)

    def test_dump_restore_dpath_remapping(self):
        """Test dumping and restoring to and from current working directory."""

        filename = "./es_test_remap_dump.oemof"

        msg = self.es.dump(
            filename,
            consider_dpath=False,
        )

        assert filename in msg

        es = EnergySystem()
        msg = es.restore(
            filename,
            consider_dpath=False,
        )

        assert filename in msg
        assert len(es.nodes) == 2
        assert isinstance(es.node["node0"], Node)

    def test_dump_restore_default_filename(self):
        """Test dumping and restoring with default filename to custom dir."""

        default_filename = "es_dump.oemof"

        with pytest.warns(
            match="Parameter 'dpath' will be removed in a future",
        ):
            msg = self.es.dump(dpath="./")

        assert default_filename in msg

        es = EnergySystem()
        with pytest.warns(
            match="Parameter 'dpath' will be removed in a future",
        ):
            msg = es.restore(dpath="./")

        assert default_filename in msg
        assert len(es.nodes) == 2
        assert isinstance(es.node["node0"], Node)

    def test_dump_restore_dpath_filename(self):
        """Test dumping and restoring with filename and dir."""

        directory = "./"
        filename = "es_dump_dpath_filename.oemof"

        with pytest.warns(
            match="Parameter 'dpath' will be removed in a future",
        ):
            msg = self.es.dump(directory, filename)

        assert filename in msg

        es = EnergySystem()
        with pytest.warns(
            match="Parameter 'dpath' will be removed in a future",
        ):
            msg = es.restore(directory, filename)

        assert filename in msg
        assert len(es.nodes) == 2
        assert isinstance(es.node["node0"], Node)

    def test_dump_restore_default(self):
        """Test default dumping and restoring."""

        default_filename = "es_dump.oemof"

        with pytest.warns(
            match="Default directory for oemof dumps will change",
        ):
            msg = self.es.dump()

        assert default_filename in msg

        es = EnergySystem()
        with pytest.warns(
            match="Default directory for oemof dumps will change",
        ):
            msg = es.restore()

        assert default_filename in msg
        assert len(es.nodes) == 2
        assert isinstance(es.node["node0"], Node)

    def test_dump_restore_impossible_combination(self):
        """Test default dumping and restoring."""

        m = "You set filename and dpath but told that dpath should be ignored."
        with pytest.raises(ValueError, match=m):
            self.es.dump(filename="foo.dump", dpath="/", consider_dpath=False)

        es = EnergySystem()

        with pytest.raises(ValueError, match=m):
            es.restore(filename="foo.dump", dpath="/", consider_dpath=False)


class TestsEnergySystem:
    def setup_method(self):
        self.es = EnergySystem()

    def test_add_nodes(self):
        assert not self.es.nodes

        node1 = Node(label="node1")
        self.es.add(node1)
        assert self.es.nodes
        assert node1 in self.es.nodes
        assert not self.es.flows()

        # Note that node2 is not added, but the Flow is already
        # registred. We do not assert the latter fact as this is not a
        # guaranteed functionality.
        node2 = Node(label="node2", inputs={node1: Edge()})
        assert node2 not in self.es.nodes

        # When both nodes are registred, also the Flow needs to be there.
        self.es.add(node2)
        assert node2 in self.es.nodes
        assert (node1, node2) in self.es.flows().keys()

    def test_enforce_unique_labels(self):
        node1 = Node(label="node1")
        self.es.add(node1)

        node2 = Node(label="node1")
        with pytest.raises(
            ValueError,
            match="already contains",
        ):
            self.es.add(node2)

        node2 = Node(label="node2")
        self.es.add(node2)

        node3 = Node(label="node1")
        node4 = Node(label="node2")
        with pytest.raises(
            ValueError,
            match="already contains",
        ):
            self.es.add(node3, node4)

    def test_automatically_add_subnode(self):
        subnetwork = Node(label="root")
        self.es.add(subnetwork)
        assert len(self.es.nodes) == 1

        leaf1 = subnetwork.subnode(Node, local_name="leaf1")
        subnetwork.subnode(Node, local_name="leaf2")
        assert len(self.es.nodes) == 3

        assert self.es.node[("leaf1", "root")] == leaf1

    def test_add_populated_subnetwork(self):
        subnetwork = Node(label="root")
        leaf1 = subnetwork.subnode(Node, local_name="leaf1")
        subnetwork.subnode(Node, local_name="leaf2")

        self.es.add(subnetwork)
        assert len(self.es.nodes) == 3

        assert self.es.node[("leaf1", "root")] == leaf1

    def test_add_flow_assignment(self):
        assert not self.es.nodes

        node0 = Node(label="node0")
        node1 = Node(label="node1")
        node2 = Node(label="node2", inputs={node0: Edge()})

        self.es.add(node0, node1, node2)

        assert (node0, node2) in self.es.flows().keys()
        assert (node1, node2) not in self.es.flows().keys()
        assert (node2, node1) not in self.es.flows().keys()

        node2.inputs[node1] = Edge()

        assert (node0, node2) in self.es.flows().keys()
        assert (node1, node2) in self.es.flows().keys()
        assert (node2, node1) not in self.es.flows().keys()

        node2.outputs[node1] = Edge()
        assert (node0, node2) in self.es.flows().keys()
        assert (node1, node2) in self.es.flows().keys()
        assert (node2, node1) in self.es.flows().keys()

    def test_check_method(self):
        node0 = Node(label="node0")
        node1 = Node(label="node1")
        node2 = Node(label="node2")
        node3 = Node(label="node3", inputs={node2: Edge()})
        self.es.check()  # empty, no problem

        self.es.add(node0)
        self.es.check()  # single node, no problem

        node0.outputs[node1] = Edge()
        with pytest.raises(RuntimeError, match="not part of EnergySystem"):
            self.es.check()  # node1 needs to be added

        self.es.add(node1)
        self.es.check()  # consistent graph added, no problem

        # The check method is not needed for these cases.
        # I still add them to the test for completeness.
        with pytest.raises(KeyError):
            node2.outputs[node1]  # not allowed anyway
        with pytest.raises(KeyError):
            node1.inputs[node2]  # also not allowed
        self.es.check()  # graph still consistent

        self.es.add(node2)
        with pytest.raises(RuntimeError, match="not part of EnergySystem"):
            self.es.check()  # if node 2 is present, node3 also needs to be

        self.es.add(node3)
        self.es.check()  # Now, everything is fine.

    def test_that_node_additions_are_signalled(self):
        """
        When a node gets `add`ed, a corresponding signal should be emitted.
        """
        node = Node(label="Node")

        def subscriber(sender, **kwargs):
            assert sender is node
            assert kwargs["EnergySystem"] is self.es
            subscriber.called = True

        subscriber.called = False

        EnergySystem.signals[EnergySystem.add].connect(subscriber, sender=node)
        self.es.add(node)
        assert subscriber.called, (
            "\nExpected `subscriber.called` to be `True`.\n"
            "Got {}.\n"
            "Probable reason: `subscriber` didn't get called."
        ).format(subscriber.called)

    def test_graph(self):
        fpath = Path(Path.home(), "test_graph_x345_efhu73.graphml")
        my_graph = graph.create_nx_graph(self.es)
        assert isinstance(my_graph, DiGraph)

        # make sure that test does not pass because of pre-existing file
        assert not fpath.is_file()

        # create graph file
        my_graph = graph.create_nx_graph(self.es, filename=fpath)
        assert isinstance(my_graph, DiGraph)
        assert fpath.is_file()

        # clean up (delete graph file)
        fpath.unlink()



# -*- coding: utf-8 -*-
"""This package contains the differnt types of Node for
modelling an energy system graph.

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Cord Kaldemeyer <>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>
SPDX-FileCopyrightText: Pierre-Francois Duc <pierre-francois@rl-institut.de>

SPDX-License-Identifier: MIT
"""

import warnings
from collections import deque

from .edge import Edge
from .entity import Entity
from .helpers import Inputs
from .helpers import Outputs


class QualifiedLabel(tuple):
    """Alias class to allow tuples in labels"""

    pass


class Node(Entity):
    r"""A Node of an energy system graph.

    Parameters
    ----------
    label : (See documentation of class `Entity`)
    inputs: list or dict, optional
        Either a list of this nodes' input nodes or a dictionary mapping input
        nodes to corresponding inflows (i.e. input values).
        List will be converted to dictionary with values set to None.
    outputs: list or dict, optional
        Either a list of this nodes' output nodes or a dictionary mapping
        output nodes to corresponding outflows (i.e. output values).
        List will be converted to dictionary with values set to None.

    Attributes
    ----------
    inputs: dict
        A dictionary mapping input nodes to corresponding inflows.
    outputs: dict
        A dictionary mapping output nodes to corresponding outflows.
    """

    def __init__(
        self,
        label,
        *,
        inputs=None,
        outputs=None,
        parent_node=None,
        custom_properties=None,
    ):
        super().__init__(label=label, custom_properties=custom_properties)

        self._inputs = Inputs(self)
        self._outputs = Outputs(self)
        self._in_edges = set()

        self.parent = parent_node

        if self.parent is not None:
            self._depth = self.parent.depth + 1
        else:
            self._depth = 0

        self.__subnodes = []
        self.__energy_system = None

        # TODO: Try to avoid this local `import`.
        from ..energy_system import EnergySystem

        EnergySystem.signals[EnergySystem.add].connect(
            self._add_subnodes, sender=self
        )

        if inputs is None:
            inputs = {}
        if outputs is None:
            outputs = {}

        msg = "{} {!r} of {!r} not an instance of Node but of {}."

        for i in inputs:
            if not isinstance(i, Node):
                raise ValueError(msg.format("Input", i, self, type(i)))
            self._in_edges.add(i)
            try:
                flow = inputs.get(i)
            except AttributeError:
                flow = None
            edge = Edge.from_object(flow)
            edge.input = i
            edge.output = self
        for o in outputs:
            if not isinstance(o, Node):
                raise ValueError(msg.format("Output", o, self, type(o)))
            try:
                flow = outputs.get(o)
            except AttributeError:
                flow = None
            edge = Edge.from_object(flow)
            edge.input = self
            edge.output = o

    @property
    def inputs(self):
        """dict:
        Dictionary mapping input :class:`Entities <Entity>` :obj:`n` to
        :class:`Edge`s from :obj:`n` into :obj:`self`.
        If :obj:`self` is an :class:`Edge`, returns a dict containing the
        :class:`Edge`'s single input node as the key and the flow as the value.
        """
        return self._inputs

    @property
    def outputs(self):
        """dict:
        Dictionary mapping output :class:`Entities <Entity>` :obj:`n` to
        :class:`Edges` from :obj:`self` into :obj:`n`.
        If :obj:`self` is an :class:`Edge`, returns a dict containing the
        :class:`Edge`'s single output node as the key and the flow as the
        value.
        """
        return self._outputs

    @property
    def depth(self) -> int:
        """
        The :class:`Node` instances have a depth defined
        as the depth of their parent (if any) + 1."""
        return self._depth

    @property
    def subnodes(self):
        """Subnodes of the Node

        It is deliberately provided as a tuple to prevent user to append
        subnodes other than with API methods.
        """
        return tuple([sn for sn in self.__subnodes])

    def add(self, *subnodes):
        """Add subnodes to this `Node`."""
        for subnode in subnodes:
            subnode.parent = self
            subnode._depth = self.depth + 1
            self.__subnodes.append(subnode)
            if self.__energy_system is not None:
                self.__energy_system.add(subnode)

    def subnode(self, class_, local_name, *args, **kwargs):
        """Create a subnode and add it to this `Node`.

        Create a subnode by calling `class_(label, *args, **kwargs)` and
        `append` the result to `self.__subnodes`.
        The purpose of this wrapper is to make sure that subnodes are
        always `label`led with a unique label.
        This is useful because this allows giving the same `local_name`
        to distinct sub-`Node`s in different `Node` s.

        Parameters
        ----------
        class_: type
            The class of the subnode to create. This class must be a subclass
            of `Node`.
        local_name: hashable
            The label to use for the subnode.
        *args, **kwargs:
            Additional positional and keyword arguments that will be passed to
            the constructor of `class_` when creating the subnode.

        Returns
        -------
        :class: Node
            The newly created subnode, which is also appended to
            `self.subnodes`.


        Examples
        --------
        Create a subnode of type `Bus` with a `label` based on the the given
        `local_name`, `inputs` and `outputs` and append it to the
        `subnodes` of this `Node`.

        When
        >>> from oemof.network import Node, Edge
        >>> subnetwork = Node("subnetwork")
        >>> input = output = Node("input")
        >>> # Create a subnode of type `Node` using this convenience function
        >>> bus = subnetwork.subnode(
        ...     Node, "bus", inputs={input: Edge()}, outputs={output: Edge()}
        ... )
        """
        if isinstance(self.label, QualifiedLabel):
            label = QualifiedLabel([local_name, *self.label])
        else:
            label = QualifiedLabel([local_name, self.label])
        subnode = class_(
            label=label,
            parent_node=self,
            *args,
            **kwargs,
        )
        self.add(subnode)
        return subnode

    def _add_subnodes(self, node, **kwargs):
        """Add subnodes to an EnergySystem.

        This is meant to be used as an event callback that is called when this
        node is added to an EnergySystem, to add the child nodes to the
        EnergySystem, too.
        """
        # TODO:
        #    Explain why the `node` argument is necessary.
        if self is not node:
            raise ValueError("Call needs to be obj._add_subnodes(obj).")
        self.__energy_system = kwargs["EnergySystem"]
        deque(
            (kwargs["EnergySystem"].add(sn) for sn in self.__subnodes),
            maxlen=0,
        )


_deprecation_warning = (
    "Usage of {} is deprecated. Use oemof.network.Node instead."
)


class Bus(Node):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            _deprecation_warning.format("oemof.network.Bus"),
            FutureWarning,
        )
        super().__init__(*args, **kwargs)


class Component(Node):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            _deprecation_warning.format("oemof.network.Component"),
            FutureWarning,
        )
        super().__init__(*args, **kwargs)


class Sink(Component):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Source(Component):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Transformer(Component):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-

"""This package contains helpers used by Entities of the energy systems.

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Cord Kaldemeyer <>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>

SPDX-License-Identifier: MIT
"""

from collections import UserDict
from collections.abc import MutableMapping


class Inputs(MutableMapping):
    """A special helper to map `n1.inputs[n2]` to `n2.outputs[n1]`."""

    def __init__(self, target):
        self.target = target

    def __getitem__(self, key):
        return key.outputs.__getitem__(self.target)

    def __delitem__(self, key):
        return key.outputs.__delitem__(self.target)

    def __setitem__(self, key, value):
        return key.outputs.__setitem__(self.target, value)

    def __iter__(self):
        return iter(self.target._in_edges)

    def __len__(self):
        return self.target._in_edges.__len__()

    def __repr__(self):
        return repr(
            "<{0.__module__}.{0.__name__}: {1!r}>".format(
                type(self), dict(self)
            )
        )


class Outputs(UserDict):
    """
    Helper that intercepts modifications to update `Inputs` symmetrically.
    """

    def __init__(self, source):
        self.source = source
        super().__init__()

    def __delitem__(self, key):
        key._in_edges.remove(self.source)
        return super().__delitem__(key)

    def __setitem__(self, key, value):
        key._in_edges.add(self.source)
        return super().__setitem__(key, value)



__version__ = "0.5.1b2"

from . import energy_system
from . import graph
from . import groupings
from . import network
from .network import Bus
from .network import Component
from .network import Edge
from .network import Entity
from .network import Node
from .network import Sink
from .network import Source
from .network import Transformer

__all__ = [
    "Bus",
    "Component",
    "Edge",
    "Entity",
    "energy_system",
    "graph",
    "groupings",
    "network",
    "Node",
    "Sink",
    "Source",
    "Transformer",
]



from .edge import Edge
from .entity import Entity
from .nodes import Bus
from .nodes import Component
from .nodes import Node
from .nodes import Sink
from .nodes import Source
from .nodes import Transformer

__all__ = [
    "Bus",
    "Component",
    "Edge",
    "Entity",
    "Node",
    "Sink",
    "Source",
    "Transformer",
]



# -*- coding: utf-8 -

"""Basic tests.

This file is part of project oemof.network (github.com/oemof/oemof-network).

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Cord Kaldemeyer <>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>

SPDX-License-Identifier: MIT
"""
from itertools import chain

from oemof.network import energy_system as es
from oemof.network.groupings import Flows
from oemof.network.groupings import FlowsWithNodes
from oemof.network.network.nodes import Node


def test_flows():
    key = object()
    ensys = es.EnergySystem(groupings=[Flows(key)])
    bus = Node(label="A Bus")
    node = Node(label="A Node", inputs={bus: None}, outputs={bus: None})
    ensys.add(bus, node)
    assert ensys.groups[key] == set(
        chain(bus.inputs.values(), bus.outputs.values())
    )


def test_flows_with_nodes():
    key = object()
    ensys = es.EnergySystem(groupings=[FlowsWithNodes(key)])
    bus = Node(label="A Bus")
    node = Node(label="A Node", inputs={bus: None}, outputs={bus: None})
    ensys.add(bus, node)
    assert ensys.groups[key], {
        (bus, node, bus.outputs[node]),
        (node, bus, node.outputs[bus]),
    }



# -*- coding: utf-8 -*-

"""This package contains the class Edge used to model
energy systems.

SPDX-FileCopyrightText: Stephan Günther <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>
SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Cord Kaldemeyer <>
SPDX-FileCopyrightText: Patrik Schönfeldt <patrik.schoenfeldt@dlr.de>

SPDX-License-Identifier: MIT
"""

from collections import namedtuple
from collections.abc import Mapping

from .entity import Entity


class Edge(Entity):
    """
    :class:`Bus`es/:class:`Component`s are always connected by an
    :class:`Edge`.

    :class:`Edge`s connect a single :class:`Node` with another. They are
    directed and have a (sequence of) value(s) attached to them, so they can
    be used to represent a flow from a source/an input to a target/an output.

    Parameters
    ----------
    input, output: :class:`Bus` or :class:`Component`, optional
    flow, values: object, optional
        The (list of) object(s) representing the values flowing from this
        edge's input into its output. Note that these two names are aliases of
        each other, so `flow` and `values` are mutually exclusive.

    Note that all of these parameters are also set as attributes with the same
    name.
    """

    Label = namedtuple("EdgeLabel", ["input", "output"])

    def __init__(
        self,
        input_node=None,
        output_node=None,
        flow=None,
        values=None,
        *,
        custom_properties=None,
    ):
        if flow is not None and values is not None:
            raise ValueError(
                "\n\n`Edge`'s `flow` and `values` keyword arguments are "
                "aliases of each other,\nso they're mutually exclusive.\n"
                "You supplied:\n"
                f"    `flow`  : {flow}\n"
                f"    `values`: {values}\n"
                "Choose one."
            )
        super().__init__(
            label=Edge.Label(input_node, output_node),
            custom_properties=custom_properties,
        )
        self.values = values if values is not None else flow
        if input_node is not None and output_node is not None:
            input_node.outputs[output_node] = self

    @classmethod
    def from_object(cls, o):
        """Creates an `Edge` instance from a single object.

        This method inspects its argument and does something different
        depending on various cases:

          * If `o` is an instance of `Edge`, `o` is returned unchanged.
          * If `o` is a `Mapping`, the instance is created by calling
            `cls(**o)`,
          * In all other cases, `o` will be used as the `values` keyword
            argument to `Edge`'s constructor.
        """
        if isinstance(o, Edge):
            return o
        elif isinstance(o, Mapping):
            return cls(**o)
        else:
            return Edge(values=o)

    @property
    def flow(self):
        return self.values

    @flow.setter
    def flow(self, values):
        self.values = values

    @property
    def input(self):
        return self.label.input

    @input.setter
    def input(self, i):
        old_input = self.input
        self._label = Edge.Label(i, self.label.output)
        if old_input is None and i is not None and self.output is not None:
            i.outputs[self.output] = self

    @property
    def output(self):
        return self.label.output

    @output.setter
    def output(self, o):
        old_output = self.output
        self._label = Edge.Label(self.label.input, o)
        if old_output is None and o is not None and self.input is not None:
            o.inputs[self.input] = self



"""Specific tests for the `oemof.groupings` module.

Most parts of the `groupings` module are tested via other tests, but certain
code paths don't get covered by those, which is what this module is for.

This file is part of project oemof (github.com/oemof/oemof). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location oemof/tests/tests_groupings.py

SPDX-License-Identifier: MIT
"""

from types import MappingProxyType as MaProTy

import pytest

from oemof.network.groupings import Grouping


def test_initialization_argument_checks():
    """`Grouping` constructor should raise `TypeError` on bad arguments."""

    message = "Grouping constructor missing required argument"
    with pytest.raises(TypeError, match=message):
        Grouping()

    message = (
        "Grouping arguments `key` and `constant_key` are  mutually exclusive."
    )
    with pytest.raises(TypeError, match=message):
        Grouping(key=lambda x: x, constant_key="key")


def test_notimplementederrors():
    """`Grouping` should raise an error when reaching unreachable code."""

    message = "There is no default implementation for `Groupings.key`."
    with pytest.raises(NotImplementedError, match=message):
        g = Grouping(key="key")
        del g.key
        g.key("dummy argument")

    message = "`Groupings.filter` called without being overridden."
    with pytest.raises(NotImplementedError, match=message):
        g = Grouping(key="key")
        del g.filter
        g.filter("dummy argument")


def test_mutable_mapping_groups():
    g = Grouping(
        key=lambda x: len(x),
        value=lambda x: {y: len([z for z in x if z == y]) for y in x},
    )
    groups = {}
    expected = {3: {"o": 2, "f": 1}}
    g("foo", groups)
    assert groups == expected, "\n  Expected: {} \n  Got     : {}".format(
        expected, groups
    )


def test_immutable_mapping_groups():
    g = Grouping(
        key=lambda x: len(x),
        value=lambda x: MaProTy({y: len([z for z in x if z == y]) for y in x}),
    )
    groups = {}
    expected = {3: MaProTy({"o": 2, "f": 1})}
    g("foo", groups)
    assert groups == expected, "\n  Expected: {} \n  Got     : {}".format(
        expected, groups
    )



# -*- coding: utf-8 -*-

"""Modules for creating and analysing energy system graphs.

This file is part of project oemof (github.com/oemof/oemof). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location oemof/oemof/graph.py

SPDX-FileCopyrightText: Simon Hilpert <>
SPDX-FileCopyrightText: Uwe Krien <uwe.krien@ifam.fraunhofer.de>

SPDX-License-Identifier: MIT
"""

import warnings
from pathlib import Path

import networkx as nx


def create_nx_graph(
    energy_system=None,
    remove_nodes=None,
    filename=None,
    remove_nodes_with_substrings=None,
    remove_edges=None,
):
    """
    Create a `networkx.DiGraph` for the passed energy system and plot it.
    See https://networkx.org/documentation/ for more information.

    Parameters
    ----------
    energy_system : `oemof.solph.network.EnergySystem`

    filename : str or Path
        Absolute filename (with path) to write your graph in the graphml
        format. If no filename is given no file will be written.

    remove_nodes: list of strings
        Nodes to be removed e.g. ['node1', node2')]

    remove_nodes_with_substrings: list of strings
        Nodes that contain substrings to be removed e.g. ['elec', 'heat')]

    remove_edges: list of string tuples
        Edges to be removed e.g. [('resource_gas', 'gas_balance')]

    Examples
    --------
    >>> import os
    >>> import pandas as pd
    >>> from oemof.network.network import Node
    >>> from oemof.network.energy_system import EnergySystem
    >>> import oemof.network.graph as graph
    >>> datetimeindex = pd.date_range('1/1/2017', periods=3, freq='h')
    >>> es = EnergySystem(timeindex=datetimeindex)
    >>> b_gas = Node(label='b_gas')
    >>> bel1 = Node(label='bel1')
    >>> bel2 = Node(label='bel2')
    >>> demand_el = Node(label='demand_el', inputs = [bel1])
    >>> pp_gas = Node(label=('pp', 'gas'),
    ...               inputs=[b_gas],
    ...               outputs=[bel1])
    >>> line_to2 = Node(label='line_to2', inputs=[bel1], outputs=[bel2])
    >>> line_from2 = Node(label='line_from2',
    ...                   inputs=[bel2], outputs=[bel1])
    >>> es.add(b_gas, bel1, demand_el, pp_gas, bel2, line_to2, line_from2)
    >>> my_graph = graph.create_nx_graph(es)
    >>> # export graph as .graphml for programs like Yed where it can be
    >>> # sorted and customized. this is especially helpful for large graphs
    >>> # grph.create_nx_graph(es, filename="my_graph.graphml")
    >>> [my_graph.has_node(nd)
    ...  for nd in ['b_gas', 'bel1', "('pp', 'gas')", 'demand_el', 'tester']]
    [True, True, True, True, False]
    >>> list(nx.attracting_components(my_graph))
    [{'demand_el'}]
    >>> sorted(list(nx.strongly_connected_components(my_graph))[1])
    ['bel1', 'bel2', 'line_from2', 'line_to2']
    >>> new_graph = graph.create_nx_graph(energy_system=es,
    ...                                  remove_nodes_with_substrings=['b_'],
    ...                                  remove_nodes=["('pp', 'gas')"],
    ...                                  remove_edges=[('bel2', 'line_from2')],
    ...                                  filename='test_graph')
    >>> [new_graph.has_node(nd)
    ...  for nd in ['b_gas', 'bel1', "('pp', 'gas')", 'demand_el', 'tester']]
    [False, True, False, True, False]
    >>> my_graph.has_edge("('pp', 'gas')", 'bel1')
    True
    >>> new_graph.has_edge('bel2', 'line_from2')
    False
    >>> os.remove('test_graph.graphml')

    Notes
    -----
    Needs graphviz and networkx (>= v.1.11) to work properly.
    Tested on Ubuntu 16.04 x64 and solydxk (debian 9).
    """
    with warnings.catch_warnings():
        # suppress ExperimentalFeatureWarnungs
        warnings.simplefilter("ignore")

        # construct graph from nodes and flows
        grph = nx.DiGraph()

        # add nodes
        for label in energy_system.node.keys():
            grph.add_node(str(label), label=str(label))

        # add labeled flows on directed edge if an optimization_model has been
        # passed or undirected edge otherwise
        for n in energy_system.nodes:
            for i in n.inputs.keys():
                weight = getattr(
                    energy_system.flows()[(i, n)], "nominal_value", None
                )
                if weight is None:
                    grph.add_edge(str(i.label), str(n.label))
                else:
                    grph.add_edge(
                        str(i.label),
                        str(n.label),
                        weight=format(weight, ".2f"),
                    )

        # remove nodes and edges based on precise labels
        if remove_nodes is not None:
            grph.remove_nodes_from(remove_nodes)
        if remove_edges is not None:
            grph.remove_edges_from(remove_edges)

        # remove nodes based on substrings
        if remove_nodes_with_substrings is not None:
            for i in remove_nodes_with_substrings:
                remove_nodes = [
                    str(label)
                    for label in energy_system.node.keys()
                    if i in str(label)
                ]
                grph.remove_nodes_from(remove_nodes)

        if filename is not None:
            nx.write_graphml(grph, Path(filename).with_suffix(".graphml"))
        return grph
